#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Tools to upload/retrieve zenodo deposits"""
import requests
import os
import sys
import json
from configparser import NoSectionError, NoOptionError

from damona.registry import ImageName, Software
from damona import Config
import colorlog

logger = colorlog.getLogger(__name__)


logger.setLevel("INFO")


class Zenodo:  # pragma: no cover
    """

    ::

        z = Zenodo(mode="sandbox.zenodo")

    You can retrieve an existing deposit (read-only) given its ID::

        deposit = z.get_deposition(959590)
            oeturn "registry.yaml"

    Or start with a new one::

        deposit = z.create_new_deposition()


    If you want to create a new file deposition and publish it, follow those 4 commands:

    1. create a new deposit::

        deposit = z.create_new_deposition()
        ID = deposit['id']
        # e.g. 959590

    2. upload data::

        up = z.upload('file', deposit)
        # this gives us the md5sum for later

    3. metadata. Here, we update the deposit with some metadata::

        data = z.get_metadata("fastqc", "v0.11.8")
        update = z.update_metadata(deposit, data)

    4. publish  NOTE THAT PUBLISHED DEPOSITION CANNOT BE MODIFIED NOR DELETED::

        published = z.publish(deposit)

    # if we want an update, we must create a new version from a published deposit::

        new_version = z.create_new_deposit_version(deposit)


    """

    def __init__(self, mode="sandbox.zenodo", token=None, author=None, affiliation=None, orcid=None):


        assert mode in ["zenodo", "sandbox.zenodo"]
        self.mode = mode
        self.headers = {"Content-Type": "application/json"}
        self.last_requests = []

        cfg = Config()

        if token is None:
            try:
                self.token = cfg.config.get(f"{mode}", "token")
                logger.info(f"Found token for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                logger.error("A token must be provided on command line or in your damona.cfg file")
                sys.exit(1)
        else:
            self.token = token

        # Do we have an orcid (optional)
        if orcid is None:
            try:
                self.orcid = cfg.config.get(f"{mode}", "orcid")
                logger.info(f"Found ORCID for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.orcid = orcid

        # Do we have a name ?
        if author is None:
            try:
                self.author = cfg.config.get(f"{mode}", "name").replace("'", "").replace('"', "")
                logger.info(f"Found Name {self.author} for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.author = author

        # Do we have an affiliation ?
        if affiliation is None:
            try:
                self.affiliation = cfg.config.get(f"{mode}", "affiliation").replace("'", "").replace('"', "")
                logger.info(f"Found Affiliation {self.affiliation} for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.affiliation = affiliation

        if self.affiliation is None or self.author is None:
            logger.error("you must provide an author and affiliation")
            sys.exit(1)

    def _get_params(self):
        return {"access_token": self.token}

    params = property(_get_params)

    def _status(self, r, correct_codes):
        self.last_requests.append(r)
        if r.status_code not in correct_codes:
            print(f"Status code: {r.status_code}; {r.reason}")
            print(f"{r.json()}")

    def create_new_deposition(self):  # pragma: no cover
        # expected status is 201

        logger.info("Creating a new deposit")
        url = f"https://{self.mode}.org/api/deposit/depositions"
        r = requests.post(url, params=self.params, json={}, headers=self.headers)
        self._status(r, [201])

        try:
            ID = self.last_requests.json()["id"]
            logger.info(f"Created a new deposit with ID={ID}")
        except:
            pass
        return r

    def get_all_depositions(self):  # pragma: no cover
        """Return all deposition for the currently authenticated user"""
        url = f"https://{self.mode}.org/api/deposit/depositions"
        r = requests.get(url, params=self.params, json={})
        self._status(r, [200])
        return r

    def get_deposition(self, ID):  # pragma: no cover
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.get(url, params=self.params, json={})
        self._status(r, [200])
        return r

    def delete_deposition(self, ID):  # pragma: no cover
        """only unpublished deposit can be deleted.

        Normal status code is 204. If already deleted, sent 401
        """
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.delete(url, params=self.params, json={})
        self._status(r, [201])
        return r

    def upload(self, filename, json_deposit):  # pragma: no cover


        try:
            bucket_url = json_deposit["links"]["bucket"]
        except:
            bucket_url = json_deposit.json()["links"]["bucket"]

        jsons = {}

        logger.info("Uploading file")
        # the upload it self may take a while
        with open(filename, "rb") as fp:
            basename = os.path.basename(filename)
            r = requests.put(f"{bucket_url}/{basename}", data=fp, params=self.params)
            self._status(r, [200])
            return r

    def get_metadata(self, software, version):
        # check validity of the version
        assert len(version.split(".")) == 3
        items = version.split(".")
        if items[0][0] != "v":
            version = f"v{version}"

        data = {
            "metadata": {
                "title": f"Damona singularity image of {software} software",
                "upload_type": "physicalobject",
                "description": f"""Singularity image(s) of {software} software to be used and installed with damona
(<a href="https://damona.readthedocs.org">See https://damona.readthedocs.org</a>) for reproducible bioinformatics
analysis.""",
                "keywords": ["apptainer", "singularity", "damona", "bioinformatics", "reproducibility", "container"],
                "version": f"{version}",
                "communities": [{"identifier": "damona"}]
            }
        }

        if self.orcid:
            data["metadata"]["creators"] = [{"orcid": self.orcid, "name": self.author, "affiliation": self.affiliation}]
        else:
            data["metadata"]["creators"] = [{"name": self.author, "affiliation": self.affiliation}]

        return data

    def update_metadata(self, data, deposit):  # pragma: no cover
        logger.info("Uploading metadata")
        ID = self.get_id(deposit)
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.put(url, params=self.params, data=json.dumps(data), headers=self.headers)
        self._status(r, [200])
        return r

    def publish(self, deposit):  # pragma: no cover
        ID = self.get_id(deposit)
        logger.info(f"Publishing {ID}")
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/publish"
        r = requests.post(url, params=self.params)
        self._status(r, [202])
        return r

    def unlock(self, deposit):  # pragma: no cover
        ID = self.get_id(deposit)
        logger.info(f"Unlocking {ID}")
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/edit"
        r = requests.post(url, params=self.params)
        self._status(r, [201, 403])
        return r

    def create_new_deposit_version(self, deposit):  # pragma: no cover
        # expected status 201
        logger.info("Creating a new version. Please wait")
        ID = self.get_id(deposit)
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/newversion"
        r = requests.post(url, params=self.params)
        self._status(r, [201])
        return r

    def get_id(self, data):
        """
        data can be the output of a requests, in which case, we call to_json() and expect a 'id'
        key. Otherwise, it is alrady a json structure. If none works, most probably just an ID (integer).
        """
        try:
            return data.json()["id"]
        except:
            try:
                return data["id"]
            except:
                return data

    def _get_registry(self):
        print(self.mode)

        if self.mode == "zenodo":
            return "registry.yaml"
        elif self.mode == "sandbox.zenodo":
            return "registry_sandbox.yaml"
    registry_name = property(_get_registry)

    def _upload(self, filename):
        data = ImageName(filename)
        software = Software(data.name)

        if software.name:
            msg = self.create_new_version_with_file_and_publish(filename)
            print(msg)
            with open(self.registry_name, "a+") as fout:
                fout.write(msg)
        else:
            msg = self.create_new_deposit_with_file_and_publish(filename)
            print(msg)
            with open(self.registry_name, "w") as fout:
                fout.write(msg)

    def create_new_deposit_with_file_and_publish(self, filename):  # pragma: no cover
        """ """
        data = ImageName(filename)

        logger.info(f"Creating and Publising new deposit for {data.name}_{data.version}")
        deposit = self.create_new_deposition()

        up = self.upload(filename, deposit)
        metadata = self.get_metadata(data.name, data.version)
        self.update_metadata(metadata, deposit)

        # in order to obtain the DOI, we must publish first
        self.publish(deposit)

        if self.last_requests[-1].ok:
            json = self.last_requests[-1].json()
            msg = self._print_info_new_deposit(data, json)
            return msg

    def _print_info_new_deposit(self, data, json):
        # figure out the filename entryy we have just added
        entry = [x for x in json["files"] if x["filename"] == data.basename][0]

        zenodo_id = json["id"]
        doi = json["conceptdoi"]
        msg = f"{data.name}:\n"
        msg += f"  doi: {doi}\n"
        msg += f"  zenodo_id: {zenodo_id}\n"
        msg += "  releases:\n"

        this_doi = json["doi"]
        record_html = json["links"]["record_html"]

        md5sum = entry["checksum"]
        filesize = entry["filesize"]
        basename = entry["filename"]
        msg += f"    {data.version}:\n"
        msg += f"      download: {record_html}/files/{basename}\n"
        msg += f"      md5sum: {md5sum}\n"
        msg += f"      doi: {this_doi}\n"
        msg += f"      filesize: {filesize}\n"  # no EOF
        return msg

    def create_new_version_with_file_and_publish(self, filename, deposit=None):

        data = ImageName(filename)
        registry_data = Software(data.name)._data
        try:
            zenodo_id = registry_data[data.name]["zenodo_id"]
        except KeyError:
            logger.critical(
                f"Could not find {data.name} in the registry. Filename must have the same name as those to ne found in the recipes."
            )
            sys.exit(1)

        logger.info(f"Creating and Publising new version for {data.name}_{data.version}")

        deposit = self.create_new_deposit_version(zenodo_id)

        # switch the state from 'done' to 'inprogress'
        self.unlock(deposit)

        # get the ID of the new latest draft and its deposition
        new_ID = deposit.json()["links"]["latest_draft"].split("/")[-1]
        new_deposit = self.get_deposition(new_ID)

        up = self.upload(filename, new_deposit.json())

        metadata = self.get_metadata(data.name, data.version)
        self.update_metadata(metadata, new_deposit)

        self.publish(new_ID)

        if self.last_requests[-1].ok:
            json = self.last_requests[-1].json()
            msg = self._print_info_new_version(data, json)
            return msg

    def _print_info_new_version(self, data, json):
        # figure out the filename entryy we have just added
        entry = [x for x in json["files"] if x["filename"] == data.basename][0]

        # update the new zenodo ID

        this_doi = json["doi"]
        md5sum = entry["checksum"]
        basename = entry["filename"]
        record_html = json["links"]["record_html"]
        filesize = entry["filesize"]

        msg = f"    {data.version}:\n"
        msg += f"      download: {record_html}/files/{basename}\n"
        msg += f"      md5sum: {md5sum}\n"
        msg += f"      doi: {this_doi}\n"
        msg += f"      filesize: {filesize}"  # no EOF
        return msg


def get_stats_id(ID):
    """Returns number of downloads"""
    from bs4 import BeautifulSoup

    r = requests.get(f"https://zenodo.org/record/{ID}")
    bs = BeautifulSoup(r.content, features="html.parser")

    for x in bs.find(id="collapse-stats").find_all("tr"):
        entries = []
        for xx in x.find_all("td"):
            entries.append(xx)
        if len(entries):
            if entries[0].contents[0].strip() == "Downloads":
                stats = entries[2].contents[0]
                return int(stats)

    # if could not find the info, return -1
    return -1



#
#  This file is part of Damona software
#
#  Copyright (c) 2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/damona/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Registry and Software manager"""
import pathlib
import glob
import sys
import os
import yaml
from yaml import Loader, CSafeLoader
import packaging.version


from damona.config import Config

import colorlog

logger = colorlog.getLogger(__name__)


__all__ = ["Releases", "Software", "Registry", "Release", "ImageName", "RemoteRegistry" ]


class ImageName:
    """Check image name

    An image name must have the following convention::

        name_x.y.z.img

    or::

        name_other_x.y.z.img

    """

    def __init__(self, name):

        self.filename = name
        self.basename = os.path.basename(name)

        if not self.basename.endswith(".img"):
            logger.error("Input file name must follow the convention NAME_x.y.z.img You provided {self.basename}")
            raise NameError

        if "_" not in self.basename:
            logger.error("Input file name must follow the convention NAME_x.y.z.img You provided {self.basename}")
            raise NameError

        name, version = self.basename.rsplit(".", 1)[0].rsplit("_", 1)

        if len(version.split(".")) != 3:
            logger.error(f"Version must be formatted as X.Y.Z. You provided {version}")
            sys.exit(1)

        self.name = name
        self.version = version

        # check version
        if len(version.split(".")) not in [2, 3]:
            raise NameError


class Releases(dict):
    """A collection of :class:`Release` for a given software.


    ::

        from damona import Software
        s = Software("fastqc")
        s.releases.last_release
    """

    def __init__(self, data):
        # a collection of releases
        self._name = list(data.keys())[0]

        if "releases" not in data[self._name]:
            pass
        else:
            for version, release in data[self._name]["releases"].items():
                # enforce the keys to be strings
                # this is for special case of x.y that are read as float instead of
                # strings for e.g x.y.z
                self[str(version)] = Release(version, data)

    def _get_last_release(self):
        from packaging import version

        # split on - for the special cases 0.11.9-py3
        return max(list(self.keys()), key=lambda x: version.parse(x.split('-')[0]))

    last_release = property(_get_last_release, doc="return the last version")



class Release:
    """A Release class

    This class populates information found in the release section of a
    registry.

    A regitry looks like::

        fastqc:
            binaries: # this is the main_binaries
            x.y.z:      # a version
                download:
                md5sum:
                binaries:  # this is the extra binaries
                exclude_binaries: # exclude binaries found in the main_binaries section

    Here we have one release named x.y.z that contains information about the
    md5sum of the container, its location, and the binaries that should be
    installed.

    If binaries of the software is empty, we use the one in the specific release.
    If no binaries are set, we use the name of the software. Duplicated are
    ignored of course.

    So future release may change the list of binaries. exclude_binaries is set
    can be used to remove such binaries

    binaries can be separated by commas or spaces.

    """

    def __init__(self, version, data):
        """

        :param version: a valid x.y.z version to be found in data[name]['release']
            data is a dictionary
        :param data:

        """
        self._name = list(data.keys())[0]

        try:
            self._binaries = self.split_binaries(data[self._name].get("binaries", []))
            kwargs = data[self._name]["releases"][version]
            self.download = kwargs["download"]
        except Exception as err:  # pragma: no cover
            logger.error(f"Incorrect formatted registry for {self._name}" + str(err))
            sys.exit(1)

        if "md5sum" not in kwargs:
            logger.debug(f"Missing md5sum entry in {self._name}. Please consider adding one ")
        if "filesize" not in kwargs:
            logger.debug(f"Missing filesize entry in {self._name}. Please consider adding one ")

        self.md5sum = kwargs.get("md5sum", None)
        self.filesize = kwargs.get("filesize", None)
        self.doi = kwargs.get("doi", None)

        self._release_binaries = self.split_binaries(kwargs.get("binaries", []))
        self._extra_binaries = self.split_binaries(kwargs.get("extra_binaries", []))
        self._exclude_binaries = self.split_binaries(kwargs.get("exclude_binaries", []))
        self._data = data

    def _get_binaries(self):
        binaries = self._binaries + self._release_binaries + self._extra_binaries
        binaries = [x for x in binaries if x not in self._exclude_binaries]
        binaries = list(set(binaries))
        binaries = sorted(list(set(binaries)))
        if len(binaries) == 0:
            return [self._name]
        else:
            return binaries

    binaries = property(_get_binaries)

    def split_binaries(self, binaries):
        if isinstance(binaries, list):
            return binaries
        else:
            return binaries.replace(",", " ").split()

    def __repr__(self):
        binaries = ",".join(self.binaries)
        txt = f"name: {self._name}\n"
        txt += f" md5: {self.md5sum}\n"
        txt += f" filesize: {self.filesize}\n"
        txt += f" doi: {self.doi}\n"
        txt += f" binaries to be installed : {binaries}\n"
        txt += f" download from: {self.download}"
        return txt


class RemoteRegistry:
    """A remote registry consists of a single yaml file


    The files is expected to be a concatenation of registry YAML files.
    See the fastqc resigtry file for an example.
    """

    def __init__(self, url):
        self.url = url
        self._read_registry()

    def _read_registry(self):

        import urllib.request

        response = urllib.request.urlopen(self.url)
        html = response.read()
        self.rawdata = html.decode("utf-8")
        remote_registry = yaml.load(self.rawdata, Loader=Loader)
        self.data = remote_registry


class BiocontainersRegistry:

    def __init__(self, filename=None):
        from damona import __path__


        if filename in [True, None]:
            self.filename = pathlib.Path(__path__[0]) / "biocontainers" /  "registry.yaml"
        elif os.path.exists(filename):
            self.filename = filename
        else:
            raise ValueError(f"Expected a valid input filename. you provided {filename}")
        self._read_registry()

    def _read_registry(self):
        # Here we use CSafeLoader to speed up reading of the large input file
        logger.info("Reading Biocontainer registry. Takes a few seconds")
        with open(self.filename, "r") as fin:
            rawdata = fin.read()
            self.data = yaml.load(rawdata, Loader=CSafeLoader)


class Software:
    """A class to read a given software registry

    A Software is made of :class:`Releases`. It contains also a name, a list of
    binaries either globally or per release, a DOI.

    ::

        registry.Software("fastqc")

    prints:

        name: fastqc
        binaries: {'0.11.9': ['fastqc'], '0.11.8': ['fastqc']}


    """

    def __init__(self, name):
        """

        :param name: a valid name to be found in the registry. Can also be a
            dictionary with expected registry format.

        """

        if isinstance(name, dict):
            # Use when calling with --url
            keys = list(name.keys())
            self.registry_name = keys[0]
            #: a :class:`Releases` attribute
            self.releases = self._interpret_registry(name)
        # we are interested in existing local YAML files
        elif os.path.exists(name) and not os.path.isdir(name):
            # print("Existing path or name")
            self.registry_name = os.path.abspath(name)
            data = self._read_registry()
            #: a :class:`Releases` attribute
            self.releases = self._interpret_registry(data)
        # directory to be found in damona directory
        else:
            from damona import __path__

            self.registry_name = pathlib.Path(__path__[0]) / "software" / name / "registry.yaml"
            data = self._read_registry()
            self._data = data
            if len(data):
                self.releases = self._interpret_registry(data)
                self.doi = data[self.name].get("doi", None)
                self.zenodo_id = data[self.name].get("zenodo_id", None)
            else:  # pragma: no cover
                self._name = None
                self._version = None

    def _read_registry(self):
        # just an alias
        regname = self.registry_name

        if os.path.exists(regname) is False:  # pragma: no cover
            logger.warning(f"Input software not found in {regname}")
            return {}

        # read the yaml

        self._yaml = yaml.load(open(regname, "r").read(), Loader=Loader)
        if len(self._yaml.keys()) != 1:  # pragma: no cover
            logger.error(f"{regname} must contain on single entry named after the images. ")
            sys.exit(1)

        return self._yaml

    def _interpret_registry(self, data):
        regname = self.registry_name
        releases = Releases(data)
        self._name = releases._name
        return releases

    def _get_name(self):
        return self._name

    name = property(_get_name)

    def _get_binaries(self):
        return dict([(rel, self.releases[rel].binaries) for rel in self.releases.keys()])

    binaries = property(_get_binaries)

    def _get_versions(self):
        return sorted(self.releases.keys())

    versions = property(_get_versions)

    def _get_md5(self):
        return dict([(rel, self.releases[rel].md5sum) for rel in self.releases.keys()])

    md5 = property(_get_md5)

    def check(self):
        pass

    def __repr__(self):
        txt = f"name: {self.name}\n"
        txt += f"binaries: {self.binaries}\n"
        txt += f"releases: {self.releases}\n"
        return txt


class Registry:
    """

    The registry contains a dictionary with all images information::

        from damona import Registry
        r = Registry()
        r.registry
        r.registry['prokka_1.14.5']['download']
        r.registry['prokka_1.14.5']['binaries']

    One can use its own registy, in which case the URL to a file called registy.txt must be provided::

        r = Registry(fro;_url="https://.../registry.txt")

    Must end in **/registry.txt**.

    """

    def __init__(self, from_url=None, biocontainers=None):
        """.. rubric:: **Constructor**

        :param str from_url: if set, uses the URL provided, otherwise uses Damona registry
        """

        self.config = Config().config
        if from_url:
            if from_url in self.config["urls"]:
                from_url = self.config["urls"][from_url]
            else:  # pragma: no cover
                assert from_url.startswith("http")
                assert from_url.endswith("registry.txt")

        self.from_biocontainers = biocontainers
        self.from_url = from_url
        self.registry = {}
        self.discovery()

    def find_candidate(self, pattern):
        """Find a unique recipe within the registry."""
        candidates = [x for x in self.registry.keys() if pattern==x or pattern in x.split(":")]

        if len(candidates) == 0:  # pragma: no cover
            logger.critical(
                f"No image found for {pattern}. Make sure it is correct. You can use 'damona search' command"
            )
            return None

        if len(candidates) == 1:
            return candidates[0]

        # sequana_tools_0.9.0 should return sequana_tools for the name and
        # 0.9.0 for the version hence the rsplit
        # similarly fastq_0.11.0-py3 should return 0.11.0

        names = [x.rsplit(":", 1)[0] for x in candidates]
        versions = [x.rsplit(":", 1)[1] for x in candidates]
        versions = [x.split("-")[0] for x in versions]
        version = max([packaging.version.parse(ver) for ver in versions])
        name = names[0]
        registry_name = pattern + ":" + str(version)

        return registry_name

    def discovery(self):
        """Look for software/release in the registry and populate the attributes"""
        if self.from_url:
            self._url_discovery()
        elif self.from_biocontainers:
            self._biocontainers_discovery()
        else:
            self._damona_discovery()

    def _populate(self, data):
        #Used by _url_discovery, _biocontainers_discovery, _damona_discovery 

        self.registry = {}

        for name, content in data.items():
            software = Software({name: content})
            software.check()
            # we may have several releases
            for version in software.versions:
                name_version = software.name + ":" + version
                release = software.releases[version]
                if name_version not in self.registry:
                    if release.download is None:  # pragma: no cover
                        logger.warning(f"software {software.name} has no download entry. please fill asap")
                    elif release.download.startswith("damona::"):  # pragma: no cover
                        from_url = self.config["urls"]["damona"]
                        release.download = release.download.replace("damona::", from_url)
                        release.download = release.download.replace("registry.txt", "")
                    self.registry[name_version] = release
                else:  # pragma: no cover
                    for kk, vv in self.registry.items():
                        print("{}: {}".format(kk, vv))
                        for kkk, vvv in self.registry[kk].items():
                            print(" - {}:  {}".format(kkk, vvv))

    def _url_discovery(self):
        ext_reg = RemoteRegistry(self.from_url)
        self._populate(ext_reg.data)

    def _biocontainers_discovery(self):

        # we read the biocontainers information
        ext_reg = BiocontainersRegistry(self.from_biocontainers)
        self._populate(ext_reg.data)

    def _damona_discovery(self):

        # read all damona registry and store in expected dictionary structure
        from damona.software import __path__
        _registry_files = glob.glob(__path__[0] + "/*/registry.yaml")
        data = {}
        for registry in _registry_files:
            software = Software(registry)
            data[software.name] = software._yaml[software.name]

        self._populate(data)


    def get_list(self, pattern=None):
        """Return list of :class:`Software` found in the registry"""
        software = {}
        for name, info in self.registry.items():
            if pattern:
                if pattern.lower() in name.lower():
                    software[name] = info.download
            else:
                software[name] = info.download
        recipes = sorted(software)
        return recipes

    def get_binaries(self, pattern=None):
        """Return binaries found and from which recipe"""
        recipes = {}
        for name, info in self.registry.items():
            if pattern:
                if pattern.lower() in [x.lower() for x in info.binaries]:
                    recipes[name] = [x for x in info.binaries if pattern.lower() in x.lower()]
            else:
                recipes[name] = info.binaries
        return recipes



#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
import pkg_resources
import os
import colorlog

try:
    version = pkg_resources.require("damona")[0].version
except Exception:  # pragma: no cover
    version = ">=0.6.0"


# The logger mechanism is here:
handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(levelname)s: [%(name)s,l %(lineno)s]: %(message)s"))
logger = colorlog.getLogger("damona")
logger.addHandler(handler)


# Here we create a persistent config directory in the Home of the user.
# This is a small file.
from damona.config import Config

try:
    Config()
except:  # pragma: no cover
    logger.warning("Could not create a persistent config file in your home. Unexpected error.")


# the following statement checks existence of environemental variables (DAMONA_PATH, DAMONA_EXE, etc)
from damona.common import DamonaInit

DamonaInit()


# The user/developer API
from damona.registry import Registry
from damona.environ import Environ, Environment
from damona.common import Damona



#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Provide some stats for admin"""
import os
import sys

from damona import Registry
from damona.registry import Software
from damona.registry import ImageName
from damona import version

import colorlog

logger = colorlog.getLogger(__name__)


__all__ = ["stats"]


def stats(biocontainers=False):
    """Prints statistics about Damona

    It includes the number of software and their releases.
    It also includes the nmbr of binaries.

    ::

        from damona.admin import stats
        stats()

    """

    data = {}
    print(f"Damona version : {version}")
    r = Registry(biocontainers=biocontainers)

    Nv = len(r.get_list())
    Nb = sum([len(x) for x in r.get_binaries().values()])
    Ns = len(set([x.split(":")[0] for x in r.get_list()]))
    if biocontainers:
        print("From biocontainers, in theory*, we also have:")
    print(f"- number of software:  {Ns}")
    print(f"- version: {Nv}")
    print(f"- unique binaries: {Nb}")
    data["version"] = Nv
    data["software"] = Ns
    data["unique_binaries"] = Nb
    if biocontainers:
        print("""*: not all software provided in the biocontainers registry are actually on Docker.
There is nothing we can do about that in Damona. Actual number is more around 1000 public software""")

    return data


def build_biocontainers_registry(output="biocontainers.yml", force=False, limit=20000): #pragma: no cover
    """Create the list of software and their versions available in Biocontainer"""
    logger.info("Retrieve all information from Biocontainers")
    try:
        from bioservices import Biocontainers
    except (ModuleNotFoundError, ImportError) as err:
        logger.error("This function is for admin only. You may use but you mut install bioservices first (pip install bioservices). ")
        return 

    b = Biocontainers()
    info = b.get_tools(limit=limit)
    if len(info) > limit:
        logger.warning(f"Looks like you reached the limit of {limit} tool. Use limit argument to get more")
    tools = {}
    for name, versions in zip(info['name'], info['versions']):
        tools[name] = [x['meta_version'] for x in versions]

    # Create the registry
    if os.path.exists(output) and not force:
        logger.error(f"Output file {output} exists already. Please rename or remove the target file")
        sys.exit(1)

    with open(output, "w") as fout:
        for name, versions in tools.items():
            fout.write(f"{name}:\n  releases:\n")
            for version in versions:
                fout.write(f"    {version}:\n")
                fout.write(f"      download: docker://biocontainers/{name}:{version}\n")




import sys
import os
from setuptools import setup, find_packages
import glob

_MAJOR               = 0
_MINOR               = 8
_MICRO               = 3
version              = '%d.%d.%d' % (_MAJOR, _MINOR, _MICRO)
release              = '%d.%d' % (_MAJOR, _MINOR)


metainfo = {
    'authors': {"main": ("Thomas Cokelaer", "thomas.cokelaer@pasteur.fr")},
    'maintainer': {"main": ("Thomas Cokelaer", "thomas.cokelaer@pasteur.fr")},
    'version': version,
    'license' : 'new BSD',
    'download_url': "https://github.com/sequana/sequana/archive/{0}.tar.gz".format(version),
    'url' : "http://github.com/sequana/sequana",
    'description': "A set of NGS singularity recipes, built for you and easily downlable" ,
    'platforms' : ['Linux', 'Unix'],
    'keywords' : ['NGS', 'singularity'],
    'classifiers' : [
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Physics']
    }


packages = find_packages()
packages = [this for this in packages if this.startswith('test.') is False]
packages = [this for this in packages if this not in ['test']]

requirements = open("requirements.txt").read().split()


from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


def copyfile():
    try:
        # This is to make sure users get the newest version installed
        # If it fails, this will be copied again when calling damona for the
        # first time.
        from easydev import CustomConfig
        configuration = CustomConfig("damona", verbose=True)
        damona_config_path = configuration.user_config_dir
        with open("damona/shell/damona.sh", "r") as fin:
            with open(damona_config_path + os.sep + "damona.sh", "w") as fout:
                fout.write(fin.read())
    except:
        # could not copy the file, we will do it when starting damona for the
        # first time. 
        pass


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        copyfile()

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        copyfile()


setup(
    name             = "damona",
    version          = version,
    maintainer       = metainfo['authors']['main'][0],
    maintainer_email = metainfo['authors']['main'][1],
    author           = metainfo['authors']['main'][0],
    author_email     = metainfo['authors']['main'][1],
    long_description = open("README.rst").read(),
    keywords         = metainfo['keywords'],
    description      = metainfo['description'],
    license          = metainfo['license'],
    platforms        = metainfo['platforms'],
    url              = metainfo['url'],
    download_url     = metainfo['download_url'],
    classifiers      = metainfo['classifiers'],

    # package installation
    packages = packages,
    install_requires = requirements,
    # specific packages for testing
    extras_require={
         "testing": [
             "pytest",
             "pytest-cov",
             "pytest-xdist",
             "pytest-mock",
             "pytest-timeout",
             "pytest-runner",
             "coveralls",
             "mock"
         ],
        "doc": [
            "sphinx>=3",
            "sphinx_rtd_theme"
            "sphinx-click"
        ]
    },

    # This is recursive include of data files
    exclude_package_data = {"": ["__pycache__"]},
    package_data = {
        'damona': ['*.cfg'],
        'damona.shell': ['*/*'],
        'damona.software' :      ['*/Singularity.*', '*/registry.yaml'],
        'damona.library' :       ['*/Singularity.*', '*/registry.yaml'],
        'damona.papers' :         ['*/registry.yaml'],
        'damona.biocontainers' : ['registry.yaml'],
        '': ["damona/shell/bash/damona.sh", "damona/shell/fish/damona.fish"]
        },

    cmdclass={"develop": PostDevelopCommand, "install": PostInstallCommand},

    zip_safe=False,
    entry_points = {
        'console_scripts':[
           'damona=damona.script:main', 
        ]
    },


)
